# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.keras.estimator namespace.
"""

import sys as _sys

from keras.estimator import model_to_estimator_v2 as model_to_estimator