# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.keras.regularizers namespace.
"""

import sys as _sys

from keras.regularizers import L1
from keras.regularizers import L1 as l1
from keras.regularizers import L1L2
from keras.regularizers import L2
from keras.regularizers import L2 as l2
from keras.regularizers import Regularizer
from keras.regularizers import deserialize
from keras.regularizers import get
from keras.regularizers import l1_l2
from keras.regularizers import serialize