# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.keras.mixed_precision namespace.
"""

import sys as _sys

from keras.api._v2.keras.mixed_precision import experimental
from keras.mixed_precision.loss_scale_optimizer import LossScaleOptimizer
from keras.mixed_precision.policy import Policy
from keras.mixed_precision.policy import global_policy
from keras.mixed_precision.policy import set_global_policy